/*
 * Name: main.c
 * Project: vuc-demo
 * License: GNU GPL v2 (see License.txt)
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include <stdint.h>
#include <string.h>

#include "main.h"
#include "uart.h"


#define BUF_SIZE	16

u8 state, buf_i, buf[BUF_SIZE];
u16 * buf16= (u16 *)buf;
vuc_adr_t adr;

u8 scope[SCOPE_SIZE], sc_i= 0, sc_state= 0, sc_last= 0;

vuc_put_meta_t put_meta[VUC_PUT_ADR_MAX+1]= {
  { 4, put_timer },
};

vuc_get_meta_t get_meta[VUC_GET_ADR_MAX+1]= {
  { get_scope },
};

/* ------------------------------------------------------------------------- */

void uart_putn04 (u8 n)
{
  n&= 0x0f;
  uart_putc(n<=9 ? n+'0' : n-10+'A');
}

void uart_putn08 (u8 n)
{
  uart_putn04(n>>4);
  uart_putn04(n&0x0f);
}

void uart_putn16 (u16 n)
{
  uart_putn08(n>>8);
  uart_putn08(n&0xff);
}

void uart_putn32 (u32 n)
{
  uart_putn16(n>>16);
  uart_putn16(n&0xffff);
}

void dbg_vals (const char * c)
{
  u8 x;

  uart_putc( 0x00 | (6+strlen(c)+1+put_meta[adr].no_of_bytes*3) );
  uart_puts("# cmd:"); uart_puts(c);
  uart_puts(" ");
  for (x= 0; x<put_meta[adr].no_of_bytes; x++)
    { uart_puts(" "); uart_putn08(buf[x]); }
}

void err (char * str)
{
  uart_putc( 0x00 | (6+strlen(str)) );
  uart_puts("!ERR: ");
  uart_puts(str);
}

void msg (const char * str, u16 data)
{
  uart_putc( 0x00 | (strlen(str)+1+4) );
  uart_puts(str);
  uart_putc(' ');
  uart_putn16(data);
}

void msg2 (const char * str, u16 data, u16 data2)
{
  uart_putc( 0x00 | (strlen(str)+4+1+4) );
  uart_puts(str);
  uart_putc(' ');
  uart_putn16(data);
  uart_putc(' ');
  uart_putn16(data2);
}

/* ------------------------------------------------------------------------- */

u8 put_timer (void)
{
  // TODO: Wait for overflow or match
  cli();
  PWM_ICR= buf16[0];
  PWM_OCR= buf16[1];
  sei();
  return (VUC_REP_STATUS<<6)|0;
}

u8 get_scope (void)
{
  msg(__FUNCTION__, 0);

  return (VUC_REP_STATUS<<6)|0;
}

/* ------------------------------------------------------------------------- */

void rep (i8 status)
{
  uart_putc(status);
  // TODO: error-codes
}

void new_state (u8 s)
{
  buf_i= 0;
  state= s;
//  msg("# state:", state);
}

u16 int_cnt= 0;

ISR(ADC_vect)
{
  u8 b= ADCH;
  ADC_TIFR|= (1<<ADC_OCF);
  if (sc_state==0) { // Wait for trigger
    if (sc_last<SCOPE_TRIGGER && b>=SCOPE_TRIGGER) {
      scope[sc_i++]= sc_last;
      scope[sc_i++]= b;
      sc_state= 1;
    }
    sc_last= b;
  } else
  if (sc_state==1) { // Triggered
    scope[sc_i++]= b;
    if (sc_i>=SCOPE_SIZE) {
      sc_i= 0;
      sc_state++; // Ready
    }
  }
  int_cnt++;
}

#define DELVAL 300000

int main (void)
{
  u16 u;
  u8 c, x;
  u32 del= DELVAL;
  char * greet= "# vuc-demo v0.0.1";
  char * ver= "vuc " VUC_PROTO_VER;

  new_state(0);

  PORTB= PORTC= PORTD= 0xff;		// activate all pull-ups
  ADC_PORT_ADR&= ~(1<<ADC_BIT_NO);	// except on ADC-Input

  DDRB= DDRC= DDRD= 0x00;		// all pins input
  LED0_DDR_ADR|= (1<<LED0_BIT_NO);	// LED-pin is output

  buf16[0]= F_CPU/8/2/1000;		// 1000 Hz
  buf16[1]= F_CPU/8/2/1000/4;		// 25% PW
  put_timer();
  // Phase and Frequency Correct PWM, TOP=ICR
  PWM_TCCR_A= (1<<PWM_COM_1);			// Clear on Match when upcounting. Set on Match when downcounting.
  PWM_TCCR_B= (1<<PWM_WGM_3)|(1<<PWM_CS_1);	// Prescaler 8
  PWM_DDR_ADR|= (1<<PWM_BIT_NO);		// PWM-pin is output

  ADC_TCCR_A= (1<<ADC_WGM_1);			// CTC Mode
  ADC_TCCR_B= (1<<ADC_CS_1)|(1<<ADC_CS_0);	// Prescaler 64
  ADC_OCR= F_CPU/64/5000;			// 5 KHz Samplerate

  ADMUX= ADC_REF_BITS|ADC_IN_BITS|(1<<ADLAR);	// Ref select, input select, Result left-adjusted
  ADCSRA= ADC_PRESC_BITS|(1<<ADEN)|(1<<ADATE)|(1<<ADIE);// Enable, Auto Trigger Enable, Int when ready
  ADCSRB= (1<<ADTS1)|(1<<ADTS0);		// Auto-trigger on timer 0 compare A

  uart_init(UART_BAUD_SELECT(BAUDRATE, F_CPU));
  sei();

  uart_putc(0x00 | strlen(greet));
  uart_puts(greet);
  uart_putc(0x00 | strlen(ver));
  uart_puts(ver);

  while (1) {
    if (sc_state==2) {
      uart_putc((VUC_REP_PUT<<6)|0);
      for (x=0; x<SCOPE_SIZE; x++)
        uart_putc(scope[x]);
      sc_state= 0;
    }
    u= uart_getc();
    if (!(u&0xff00)) {
      c= u&0xff;
      switch (state) {
        case 0:
          adr= c&0x3f; c= c>>6;
          if (c>VUC_CMD__MAX) {
            err("Bad cmd");
          } else {
            // TODO: adr above 64
            if ( (c==VUC_CMD_PUT && adr>VUC_PUT_ADR_MAX) || (c==VUC_CMD_GET && adr>VUC_GET_ADR_MAX) ) {
              err("Bad adr");
            } else {
              new_state((c+1)<<4);
            }
          }
          break;
        case (VUC_CMD_ASCII+1)<<4:
          if (buf_i<adr) {
            buf[buf_i++]= c;
          } else {
            // TODO: extended commands
            new_state(0);
          }
          break;
        case (VUC_CMD_PUT+1)<<4:
          buf[buf_i++]= c;
          if (buf_i>=put_meta[adr].no_of_bytes) {
            rep(put_meta[adr].func());
            new_state(0);
          }
          break;
        case (VUC_CMD_GET+1)<<4:
          rep(get_meta[adr].func());
          new_state(0);
          break;
      }
    }
    if (!(del--)) {
      LED0_PORT_ADR^= (1<<LED0_BIT_NO);	// Toggle LED
      if (LED0_PORT_ADR&(1<<LED0_BIT_NO)) {
//        msg("# sc_state:", sc_state);
      }
      del= DELVAL;
    }
  }

  uart_exit();
  return 0;
}

/* ------------------------------------------------------------------------- */
